----------------------------------------------------------------------
----                                                              ----
---- MC68K family compatible IP Core                              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- This model provides an opcode and bus timing compatible ip   ----
---- core compared to Motorola's MC68851 memory management unit.  ----
----                                                              ----
---- This file contains the 68851's bus interface unit.           ----
----                                                              ----
---- Description:                                                 ----
---- This module provides bus control during read cycles, write   ----
---- cycles and read modify write cycles. It also provides 2 and  ----
---- 3 wire bus arbitration control, halt and rerun operation,    ----
---- bus and address error generation, wait state insertion and   ----
---- synchronous bus operation.                                   ----
---- In the following there are some remarks on the working       ----
---- principle of this core.                                      ----
----                                                              ----
---- Bus cycle operation:                                         ----
---- A bus cycle is released by either asserting RD_BUS for       ----
---- entering a read cycle, WR_BUS for entering a write cycle.    ----
---- There must not be asserted both signals at the same time.    ----
---- Once the bus cycle is started, the RD_BUS or WR_BUS must be  ----
---- asserted until the cycle finishes. This is indicated by the  ----
---- signal BUS_CTRL_RDY, which is valid for one clock cycle      ----
---- after the bus operation finished.                            ----
----                                                              ----
---- Bus arbitration topics:                                      ----
---- No bus arbitration during read modify write cycle.           ----
---- In the case of a 2 wire bus arbitration, no re-entry for     ----
----   another device is possible.                                ----
----                                                              ----
---- Bus error topics:                                            ----
---- During a bus error, the bus cycle finishes also asserting    ----
---- the BUS_CTRL_RDY signal during S5 for read, write and the    ----
---- read portion of the read modify write cycle and during S5    ----
---- for the write portion of a read modify write cycle.          ----
----                                                              ----
---- Bus re-run topics:                                           ----
---- During a re-run condition, the bus cycle finishes also       ----
---- asserting the BUS_CTRL_RDY signal during S5 for read and     ----
---- write cycle and during S5 (write) for the read modify write  ----
---- cycle.                                                       ----
----                                                              ----
---- RESET topics:                                                ----
---- When a reset is released by the CPU due to the RESET_EN      ----
---- control, the RESET_RDY indicates the finishing of the reset  ----
---- cycle 512 clock cycles later. The RESET_EN may be asserted   ----
---- until RESET_RDY indicates 'ready'. The RESET_RDY is a strobe ----
---- and therefore valid for one clock cycle.                     ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright (C) 2007 Wolfgang Foerster                         ----
----                                                              ----
---- This source file may be used and distributed without         ----
---- restriction provided that this copyright statement is not    ----
---- removed from the file and that any derivative work contains  ----
---- the original copyright notice and the associated disclaimer. ----
----                                                              ----
---- This source file is free software; you can redistribute it   ----
---- and/or modify it under the terms of the GNU Lesser General   ----
---- Public License as published by the Free Software Foundation; ----
---- either version 2.1 of the License, or (at your option) any   ----
---- later version.                                               ----
----                                                              ----
---- This source is distributed in the hope that it will be       ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied   ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ----
---- PURPOSE. See the GNU Lesser General Public License for more  ----
---- details.                                                     ----
----                                                              ----
---- You should have received a copy of the GNU Lesser General    ----
---- Public License along with this source; if not, download it   ----
---- from http://www.gnu.org/licenses/lgpl.html                   ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K7B  2007/12/24 WF
--   Initial Release.
-- 

library work;
use work.wf68851ip_pkg.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WF68851IP_BUS_INTERFACE is
	port (
		-- System control:
		CLK					: in bit; -- System clock.
		RESETn				: in bit; -- Core reset.
		RESET_INn			: in bit; -- System's reset input.
		RESET_OUT_EN		: out bit; -- System's reset output open drain enable.
		RESET_CPUn			: out bit; -- Internal reset used for CPU initialization.
		RESET_EN			: in bit; -- Force reset control.
		RESET_RDY			: out bit; -- Indicates the end of a forced reset cycle (strobe).

		-- Data lines:
		DATA_PORT_IN		: in std_logic_vector(31 downto 0); -- Data bus input lines.
		DATA_PORT_OUT		: out std_logic_vector(31 downto 0); -- Data bus output lines.
		DATA_2_PORT			: in std_logic_vector(31 downto 0); -- Internal bus input lines.
		DATA_2_CORE			: out std_logic_vector(31 downto 0); -- Internal bus output lines.
		OPCODE_DATA			: out std_logic_vector(15 downto 0); -- Data for the operation code decoder.

		-- Receive buffers:
		SEL_BUFF_A			: in bit;
		SEL_BUFF_B			: in bit;
		BUFFER_A			: out std_logic_vector(31 downto 0); -- Receive buffer A.
		BUFFER_B			: out std_logic_vector(31 downto 0); -- Receive buffer B.

		-- Control signals:
		RD_BUS				: in bit; -- Read bus request.
		WR_BUS				: in bit; -- Wriyte request.
		A0					: in std_logic; -- Least significant bit of the address counter.
		A1					: in std_logic; -- .. and the second least.
		PC1					: in std_logic; -- Second least significant bit of the program counter.
		OP_SIZE				: in CTRL_OP_SIZETYPE; -- Used for the input multiplexer (buffers).
		OP_END				: in bit; -- Indicates the end of an operation, used for the read modify writes.
		RMC_OP				: in bit; -- Indicates wether there is a RMC in progress.
		EXEC_ABORT			: in bit; -- Skip the current bus cycle, if '1'.
		ADR_OFFS_CLR		: in bit; -- Temporary address offset clear.
		MOVEP				: in bit; -- Indicates a MOVEP operation.
		ADR_OFFSET			: out std_logic_vector(5 downto 0);
		BUS_CTRL_RDY		: out bit; -- Indicates 'Ready'.
		DATA_VALID			: out bit;

		-- Tri state controls:
		DATA_PORT_EN		: out bit;
		BUS_EN				: out bit;

		-- Bus control signals:
		DSACKn				: in bit_vector(1 downto 0);
		BERRn				: in bit;
		AVECn				: in bit;
		HALTn				: in std_logic;

		SIZE				: out std_logic_vector(1 downto 0);
		ASn					: out bit;
		DSn					: out bit;
		RWn					: out bit;
		RMCn				: out bit;

		-- Bus arbitration:
		BRn					: in bit;
		BGACKn				: in bit;
		BGn					: out bit;
		
		-- Cache controls:
		OPCODE_FETCH		: in bit;
		FLUSH_OPC_BUFF		: in bit
		);
end entity WF68851IP_BUS_INTERFACE;
	
architecture BEHAVIOR of WF68851IP_BUS_INTERFACE is
type BUS_CTRL_STATES is (IDLE, BUSY_RD, BUSY_WR);
type ARB_STATES is(IDLE, GRANT, WAIT_RELEASE_3WIRE, WAIT_RELEASE_2WIRE);
type BUS_WIDTH_TYPE is(LONG_32, WORD, BYTE);
type OPC_BUFF_STATUS_TYPE is (full, empty);
type TIME_SLICES is (IDLE, S0, S1, S2, S3, S4, S5);
signal BUS_CTRL_STATE		: BUS_CTRL_STATES;
signal NEXT_BUS_CTRL_STATE	: BUS_CTRL_STATES;
signal RESET_OUT_EN_I		: bit;
signal RESET_CPU_In			: bit;
signal ARB_STATE			: ARB_STATES;
signal NEXT_ARB_STATE		: ARB_STATES;
signal T_SLICE				: TIME_SLICES;
signal SLICE_CNT			: std_logic_vector(1 downto 0);
signal BUS_WIDTH			: BUS_WIDTH_TYPE;
signal BUS_CYC_RDY			: bit;
signal BUS_EN_I				: bit;
signal DS_RD_EN_N			: bit;
signal DS_RD_EN				: bit;
signal DS_WR_EN_N			: bit;
signal DS_WR_EN				: bit;
signal AS_EN_N				: bit;
signal AS_EN				: bit;
signal DATA_EN_P			: bit;
signal DATA_EN				: bit;
signal WAITSTATES			: bit;
signal BUS_LOCKED			: bit;
signal OPCODE_BUFFER		: std_logic_vector(31 downto 0);
signal OPC_BUFF_STATUS		: OPC_BUFF_STATUS_TYPE;
signal SIZE_I				: std_logic_vector(1 downto 0);
signal ADR_10				: std_logic_vector(1 downto 0);
signal ADR_OFFS_I			: std_logic_vector(5 downto 0);
signal RMC_LOCK				: bit;
begin
	ADR_10 <= A1 & A0;
	SIZE <= SIZE_I;

	DATA_PORT_OUT <= -- Data output multiplexer.
		-- Byte: 
		DATA_2_PORT(7 downto 0) & DATA_2_PORT(7 downto 0) & DATA_2_PORT(7 downto 0) & DATA_2_PORT(7 downto 0) when SIZE_I = "01" else
		-- Word:
		DATA_2_PORT(15 downto 0) & DATA_2_PORT(15 downto 0) when SIZE_I = "10" and A0 = '0' else
		DATA_2_PORT(15 downto 8) & DATA_2_PORT(15 downto 0) & DATA_2_PORT(15 downto 8) when SIZE_I = "10" and A0 = '1' else
		-- 3 bytes:
		DATA_2_PORT(23 downto 0) & DATA_2_PORT(31 downto 24) when SIZE_I = "11" and A1 = '0' and A0 = '0' else
		DATA_2_PORT(23 downto 16) & DATA_2_PORT(23 downto 8) & DATA_2_PORT(7 downto 0) when SIZE_I = "11" and A1 = '0' and A0 = '1' else
		DATA_2_PORT(23 downto 8) & DATA_2_PORT(23 downto 8) when SIZE_I = "11" and A1 = '1' and A0 = '0' else
		DATA_2_PORT(23 downto 16) & DATA_2_PORT(23 downto 8) & DATA_2_PORT(23 downto 16) when SIZE_I = "11" and A1 = '1' and A0 = '1' else
		-- Long words:
		DATA_2_PORT(31 downto 0) when SIZE_I = "00" and A1 = '0' and A0 = '0' else
		DATA_2_PORT(31 downto 24) & DATA_2_PORT(31 downto 8) when SIZE_I = "00" and A1 = '0' and A0 = '1' else
		DATA_2_PORT(31 downto 16) & DATA_2_PORT(31 downto 16) when SIZE_I = "00" and A1 = '1' and A0 = '0' else
		DATA_2_PORT(31 downto 24) & DATA_2_PORT(31 downto 16) & DATA_2_PORT(31 downto 24); -- SIZE_I = "00", A1 = '1', A0 = '1'.

	BUS_BUFFER: process(RESETn, CLK)
	-- This process samples the data from the data bus during the bus phase S4.
	-- During S4 the received data from the bus is valid depending on the selection
	-- of BUS_WIDTH, SIZE, A1 and A0.
	begin
		if RESETn = '0' then
			BUFFER_A <= (others => '0');
			BUFFER_B <= (others => '0');
		elsif CLK = '0' and CLK' event then
			-- Bus buffer A:
			if T_SLICE = S4 and SEL_BUFF_A = '1' then
				case BUS_WIDTH is
					when BYTE =>
						case SIZE_I is
							when "01" => -- Byte.
								BUFFER_A(7 downto 0) <= DATA_PORT_IN(31 downto 24);
							when "10" => -- Word.
								BUFFER_A(15 downto 8) <= DATA_PORT_IN(31 downto 24);
							when "11" => -- Three bytes.
								BUFFER_A(23 downto 16) <= DATA_PORT_IN(31 downto 24);
							when "00" => -- Long word.
								BUFFER_A(31 downto 24) <= DATA_PORT_IN(31 downto 24);
							when others => null;
						end case;
					when WORD =>
						case SIZE_I is
							when "01" => -- Byte.
								case ADR_10 is
									when "00" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "01" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "11" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "10" => -- Word.
								case ADR_10 is
									when "00" => BUFFER_A(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_A(15 downto 8) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_A(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "11" => BUFFER_A(15 downto 8) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "11" => -- Three bytes.
								case ADR_10 is
									when "00" => BUFFER_A(23 downto 8) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_A(23 downto 16) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_A(23 downto 8) <= DATA_PORT_IN(31 downto 16);
									when "11" => BUFFER_A(23 downto 16) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "00" => -- Long word.
								case ADR_10 is
									when "00" => BUFFER_A(31 downto 16) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_A(31 downto 24) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_A(31 downto 16) <= DATA_PORT_IN(31 downto 16);
									when "11" => BUFFER_A(31 downto 24) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when others => null;
						end case;
					when LONG_32 =>
						case SIZE_I is
							when "01" => -- Byte.
								case ADR_10 is
									when "00" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "01" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(15 downto 8);
									when "11" => BUFFER_A(7 downto 0) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "10" => -- Word.
								case ADR_10 is
									when "00" => BUFFER_A(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_A(15 downto 0) <= DATA_PORT_IN(23 downto 8);
									when "10" => BUFFER_A(15 downto 0) <= DATA_PORT_IN(15 downto 0);
									when "11" => BUFFER_A(15 downto 8) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "11" => -- Three bytes.
								case ADR_10 is
									when "00" => BUFFER_A(23 downto 0) <= DATA_PORT_IN(31 downto 8);
									when "01" => BUFFER_A(23 downto 0) <= DATA_PORT_IN(23 downto 0);
									when "10" => BUFFER_A(23 downto 8) <= DATA_PORT_IN(15 downto 0);
									when "11" => BUFFER_A(23 downto 16) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "00" => -- Long word.
								case ADR_10 is
									when "00" => BUFFER_A(31 downto 0) <= DATA_PORT_IN(31 downto 0);
									when "01" => BUFFER_A(31 downto 8) <= DATA_PORT_IN(23 downto 0);
									when "10" => BUFFER_A(31 downto 16) <= DATA_PORT_IN(15 downto 0);
									when "11" => BUFFER_A(31 downto 24) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when others => null;
						end case;
				end case;
			-- Bus buffer B:
			elsif T_SLICE = S4 and SEL_BUFF_B = '1' then
				case BUS_WIDTH is
					when BYTE =>
						case SIZE_I is
							when "01" => -- Byte.
								BUFFER_B(7 downto 0) <= DATA_PORT_IN(31 downto 24);
							when "10" => -- Word.
								BUFFER_B(15 downto 8) <= DATA_PORT_IN(31 downto 24);
							when "11" => -- Three bytes.
								BUFFER_B(23 downto 16) <= DATA_PORT_IN(31 downto 24);
							when "00" => -- Long word.
								BUFFER_B(31 downto 24) <= DATA_PORT_IN(31 downto 24);
							when others => null;
						end case;
					when WORD =>
						case SIZE_I is
							when "01" => -- Byte.
								case ADR_10 is
									when "00" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "01" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "11" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "10" => -- Word.
								case ADR_10 is
									when "00" => BUFFER_B(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_B(15 downto 8) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_B(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "11" => BUFFER_B(15 downto 8) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "11" => -- Three bytes.
								case ADR_10 is
									when "00" => BUFFER_B(23 downto 8) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_B(23 downto 16) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_B(23 downto 8) <= DATA_PORT_IN(31 downto 16);
									when "11" => BUFFER_B(23 downto 16) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "00" => -- Long word.
								case ADR_10 is
									when "00" => BUFFER_B(31 downto 16) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_B(31 downto 24) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_B(31 downto 16) <= DATA_PORT_IN(31 downto 16);
									when "11" => BUFFER_B(31 downto 24) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when others => null;
						end case;
					when LONG_32 =>
						case SIZE_I is
							when "01" => -- Byte.
								case ADR_10 is
									when "00" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "01" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when "10" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(15 downto 8);
									when "11" => BUFFER_B(7 downto 0) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "10" => -- Word.
								case ADR_10 is
									when "00" => BUFFER_B(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "01" => BUFFER_B(15 downto 0) <= DATA_PORT_IN(23 downto 8);
									when "10" => BUFFER_B(15 downto 0) <= DATA_PORT_IN(15 downto 0);
									when "11" => BUFFER_B(15 downto 8) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "11" => -- Three bytes.
								case ADR_10 is
									when "00" => BUFFER_B(23 downto 0) <= DATA_PORT_IN(31 downto 8);
									when "01" => BUFFER_B(23 downto 0) <= DATA_PORT_IN(23 downto 0);
									when "10" => BUFFER_B(23 downto 8) <= DATA_PORT_IN(15 downto 0);
									when "11" => BUFFER_B(23 downto 16) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "00" => -- Long word.
								case ADR_10 is
									when "00" => BUFFER_B(31 downto 0) <= DATA_PORT_IN(31 downto 0);
									when "01" => BUFFER_B(31 downto 8) <= DATA_PORT_IN(23 downto 0);
									when "10" => BUFFER_B(31 downto 16) <= DATA_PORT_IN(15 downto 0);
									when "11" => BUFFER_B(31 downto 24) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when others => null;
						end case;
				end case;
			end if;

			-- Core buffer:
			if T_SLICE = S4 then
				case BUS_WIDTH is
					when BYTE =>
						case SIZE_I is
							when "01" => -- Byte.
								DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(31 downto 24);
							when "10" => -- Word.
								DATA_2_CORE(15 downto 8) <= DATA_PORT_IN(31 downto 24);
							when "11" => -- Three bytes.
								DATA_2_CORE(23 downto 16) <= DATA_PORT_IN(31 downto 24);
							when "00" => -- Long word.
								DATA_2_CORE(31 downto 24) <= DATA_PORT_IN(31 downto 24);
							when others => null;
						end case;
					when WORD =>
						case SIZE_I is
							when "01" => -- Byte.
								case ADR_10 is
									when "00" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "01" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when "10" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "11" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "10" => -- Word.
								case ADR_10 is
									when "00" => DATA_2_CORE(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "01" => DATA_2_CORE(15 downto 8) <= DATA_PORT_IN(23 downto 16);
									when "10" => DATA_2_CORE(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "11" => DATA_2_CORE(15 downto 8) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "11" => -- Three bytes.
								case ADR_10 is
									when "00" => DATA_2_CORE(23 downto 8) <= DATA_PORT_IN(31 downto 16);
									when "01" => DATA_2_CORE(23 downto 16) <= DATA_PORT_IN(23 downto 16);
									when "10" => DATA_2_CORE(23 downto 8) <= DATA_PORT_IN(31 downto 16);
									when "11" => DATA_2_CORE(23 downto 16) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when "00" => -- Long word.
								case ADR_10 is
									when "00" => DATA_2_CORE(31 downto 16) <= DATA_PORT_IN(31 downto 16);
									when "01" => DATA_2_CORE(31 downto 24) <= DATA_PORT_IN(23 downto 16);
									when "10" => DATA_2_CORE(31 downto 16) <= DATA_PORT_IN(31 downto 16);
									when "11" => DATA_2_CORE(31 downto 24) <= DATA_PORT_IN(23 downto 16);
									when others => null;
								end case;
							when others => null;
						end case;
					when LONG_32 =>
						case SIZE_I is
							when "01" => -- Byte.
								case ADR_10 is
									when "00" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(31 downto 24);
									when "01" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(23 downto 16);
									when "10" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(15 downto 8);
									when "11" => DATA_2_CORE(7 downto 0) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "10" => -- Word.
								case ADR_10 is
									when "00" => DATA_2_CORE(15 downto 0) <= DATA_PORT_IN(31 downto 16);
									when "01" => DATA_2_CORE(15 downto 0) <= DATA_PORT_IN(23 downto 8);
									when "10" => DATA_2_CORE(15 downto 0) <= DATA_PORT_IN(15 downto 0);
									when "11" => DATA_2_CORE(15 downto 8) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "11" => -- Three bytes.
								case ADR_10 is
									when "00" => DATA_2_CORE(23 downto 0) <= DATA_PORT_IN(31 downto 8);
									when "01" => DATA_2_CORE(23 downto 0) <= DATA_PORT_IN(23 downto 0);
									when "10" => DATA_2_CORE(23 downto 8) <= DATA_PORT_IN(15 downto 0);
									when "11" => DATA_2_CORE(23 downto 16) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when "00" => -- Long word.
								case ADR_10 is
									when "00" => DATA_2_CORE(31 downto 0) <= DATA_PORT_IN(31 downto 0);
									when "01" => DATA_2_CORE(31 downto 8) <= DATA_PORT_IN(23 downto 0);
									when "10" => DATA_2_CORE(31 downto 16) <= DATA_PORT_IN(15 downto 0);
									when "11" => DATA_2_CORE(31 downto 24) <= DATA_PORT_IN(7 downto 0);
									when others => null;
								end case;
							when others => null;
						end case;
				end case;
			end if;
		end if;
	end process BUS_BUFFER;

	OPCODE_BUFF: process(RESETn, CLK)
	-- This opcode buffer works on long word boundaries and is intended to store
	-- two operation words or two words of a multiword operation at a time.
	begin
		if RESETn = '0' then
			OPCODE_BUFFER <= (others => '0');
		elsif CLK = '0' and CLK' event then
			if T_SLICE = S4 and OPCODE_FETCH = '1' then
				case BUS_WIDTH is
					when BYTE =>
						case SIZE_I is
							when "00" => -- Long word.
								OPCODE_BUFFER(31 downto 24) <= DATA_PORT_IN(31 downto 24);
							when "11" => -- Three bytes.
								OPCODE_BUFFER(23 downto 16) <= DATA_PORT_IN(31 downto 24);
							when "10" => -- Word.
								OPCODE_BUFFER(15 downto 8) <= DATA_PORT_IN(31 downto 24);
							when "01" => -- Byte.
								OPCODE_BUFFER(7 downto 0) <= DATA_PORT_IN(31 downto 24);
							when others => null;
						end case;
					when WORD =>
						if SIZE_I = "00" then -- Hi Word.
							OPCODE_BUFFER(31 downto 16) <= DATA_PORT_IN(31 downto 16);
						else -- Low word.
							OPCODE_BUFFER(15 downto 0) <= DATA_PORT_IN(31 downto 16);
						end if;
					when LONG_32 =>
						OPCODE_BUFFER <= DATA_PORT_IN;
				end case;
			end if;
		end if;
	end process OPCODE_BUFF;

	BUFFER_STATUS: process(RESETn, CLK)
	-- The status of the OPCODE_BUFFER is important for the switching of the OPCODE_DATA
	-- which comes form the bus or, if full, from the buffer.
	-- The buffer is declared full, when the last data packet is written into the buffer
	-- and when the program counter address PC(1) is zero, which means the opcode is aligned
	-- to long word boundaries. Otherwise, the data of the high word is discarded and the low
	-- word is written simultaneously to the fetch registers and to the buffer. In this case
	-- the buffer contains only a copy of the last instruction word and is not declared full.
	begin
		if RESETn = '0' then
			OPC_BUFF_STATUS <= empty;
		elsif CLK = '1' and CLK' event then
			if FLUSH_OPC_BUFF = '1' then
				OPC_BUFF_STATUS <= empty;
			elsif BUS_CTRL_STATE = IDLE and RD_BUS = '1' and OPCODE_FETCH = '1' and OPC_BUFF_STATUS = full then
				OPC_BUFF_STATUS <= empty;
			-- Bus access, update before the size information in S4 changes:
			elsif T_SLICE = S3 and OPCODE_FETCH = '1' then
				case BUS_WIDTH is
					when BYTE =>
						if SIZE_I = "01" and PC1 = '0' then
							OPC_BUFF_STATUS <= full; -- Filling finished after fourth bus access.
						end if;
					when WORD =>
						if SIZE_I = "10" and PC1 = '0' then
							OPC_BUFF_STATUS <= full; -- Filling finished after second bus access.
						end if;
					when LONG_32 =>
						if PC1 = '0' then
							OPC_BUFF_STATUS <= full; -- Filling finished after first bus access.
						end if;
				end case;
			end if;
		end if;
	end process BUFFER_STATUS;

	-- Opcode data Multiplexer:
	-- This multiplexer takes into account, that the entire OPCODE_BUFFER is written from a long word
	-- address boundary (PC(1 downto 0) = "00")
	OPCODE_DATA <=  DATA_PORT_IN(15 downto 0) when OPC_BUFF_STATUS = empty and PC1 = '1' and BUS_WIDTH = LONG_32 else
					DATA_PORT_IN(31 downto 16) when OPC_BUFF_STATUS = empty and PC1 = '1' and BUS_WIDTH = WORD else
					OPCODE_BUFFER(15 downto 8) & DATA_PORT_IN(31 downto 24) when OPC_BUFF_STATUS = empty and PC1 = '1' else
					OPCODE_BUFFER(31 downto 16) when OPC_BUFF_STATUS = empty and PC1 = '0' else 
					OPCODE_BUFFER(31 downto 16) when OPC_BUFF_STATUS = full and PC1 = '0' else
					OPCODE_BUFFER(15 downto 0); -- Opcode buffer is full and PC1 = '1'.

	DATA_VALID <= 	'1' when T_SLICE = S4 and A1 = '1' else -- Sample opcode data during S4, when opcode buffer run full.
					'1' when T_SLICE = IDLE and RD_BUS = '1' and OPC_BUFF_STATUS = full and OPCODE_FETCH = '1' else '0';

	BUS_CTRL_REG: process(RESETn, CLK)
	begin
		if RESETn = '0' then
			BUS_CTRL_STATE <= IDLE;
		elsif CLK = '1' and CLK' event then
			BUS_CTRL_STATE <= NEXT_BUS_CTRL_STATE;
		end if;
	end process BUS_CTRL_REG;

	BUS_CTRL_DEC: process(BUS_CTRL_STATE, RD_BUS, WR_BUS, BUS_CYC_RDY, SIZE_I, BUS_LOCKED, 
                                          OPCODE_FETCH, OPC_BUFF_STATUS, HALTn, BERRn, RESET_CPU_In, EXEC_ABORT)
	begin
		case BUS_CTRL_STATE is
			when IDLE =>
				if BUS_LOCKED = '1' or RESET_CPU_In = '0' or HALTn = '0' then
					NEXT_BUS_CTRL_STATE <= IDLE; -- Bus in arbitration, reset or halted.
				elsif RD_BUS = '1' and OPCODE_FETCH = '1' and OPC_BUFF_STATUS = full then
					NEXT_BUS_CTRL_STATE <= IDLE; -- Do not start a bus sequence when there is valid data in the buffer!
				elsif RD_BUS = '1' then
					NEXT_BUS_CTRL_STATE <= BUSY_RD;
				elsif WR_BUS = '1' then
					NEXT_BUS_CTRL_STATE <= BUSY_WR;
				else
					NEXT_BUS_CTRL_STATE <= IDLE;
				end if;
			-- A SIZE_I count of "00" means normally four bytes to be transfered.
			-- After a bus transfer a value of "00" indicates that no further
			-- bytes are required for a bus transfer.
			when BUSY_RD =>
				if EXEC_ABORT = '1' and BUS_CYC_RDY = '1' then
					NEXT_BUS_CTRL_STATE <= IDLE;
				elsif BUS_CYC_RDY = '1' and HALTn = '0' and BERRn = '0' then -- Retry cycle.
					NEXT_BUS_CTRL_STATE <= BUSY_RD;
				elsif BUS_CYC_RDY = '1' and SIZE_I = "00" then
					NEXT_BUS_CTRL_STATE <= IDLE;
				else
					NEXT_BUS_CTRL_STATE <= BUSY_RD;
				end if;
			when BUSY_WR =>
				if EXEC_ABORT = '1' and BUS_CYC_RDY = '1' then
					NEXT_BUS_CTRL_STATE <= IDLE;
				elsif BUS_CYC_RDY = '1' and HALTn = '0' and BERRn = '0' then -- Retry cycle.
					NEXT_BUS_CTRL_STATE <= BUSY_WR;
				elsif BUS_CYC_RDY = '1' and SIZE_I = "00" then
					NEXT_BUS_CTRL_STATE <= IDLE;
				else
					NEXT_BUS_CTRL_STATE <= BUSY_WR;
				end if;
		end case;
	end process BUS_CTRL_DEC;

	P_SIZE: process(RESETn, CLK)
	-- The information about the SIZE_I remaining for a bus transfer is updated in state number 4.
	-- The initial size is sampled right before the bus acces. This requires the RD_BUS and WR_BUS
	-- signals to work on the positive clock edge.
	begin
		if RESETn = '0' then
			SIZE_I <= "00";
		elsif CLK = '0' and CLK' event then
			if BUS_CTRL_STATE = IDLE then -- Sample.
				case OP_SIZE is
					when LONG => 	SIZE_I <= "00";
					when TRIPLE =>	SIZE_I <= "11";
					when WORD => 	SIZE_I <= "10";
					when BYTE => 	SIZE_I <= "01";
				end case;
			elsif BUS_CTRL_STATE = BUSY_RD and HALTn = '0' and BERRn = '0' then
			 	null; -- Do not update if there is a retry cycle.
			elsif BUS_CTRL_STATE = BUSY_WR and HALTn = '0' and BERRn = '0' then
			 	null; -- Do not update if there is a retry cycle.
			elsif BUS_CTRL_STATE = BUSY_RD and NEXT_BUS_CTRL_STATE =  BUSY_WR then -- Sample read modify write.
				case OP_SIZE is
					when LONG => 	SIZE_I <= "00";
					when TRIPLE =>	SIZE_I <= "11";
					when WORD => 	SIZE_I <= "10";
					when BYTE => 	SIZE_I <= "01";
				end case;
			elsif T_SLICE = S4 then
				if MOVEP = '1' and OPCODE_FETCH = '0' then
					SIZE_I <= SIZE_I - '1'; -- Special case.
				--
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "00" and ADR_10 = "00" then
					SIZE_I <= "00";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "00" and ADR_10 = "01" then
					SIZE_I <= SIZE_I - "11";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "00" and ADR_10 = "10" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "00" and ADR_10 = "11" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "11" and ADR_10 = "00" then
					SIZE_I <= SIZE_I - "11";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "11" and ADR_10 = "01" then
					SIZE_I <= SIZE_I - "11";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "11" and ADR_10 = "10" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "11" and ADR_10 = "11" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "10" and ADR_10 = "00" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "10" and ADR_10 = "01" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "10" and ADR_10 = "10" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "10" and ADR_10 = "11" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = LONG_32 and SIZE_I = "01" then
					SIZE_I <= SIZE_I - "01";
				--
				elsif BUS_WIDTH = WORD and SIZE_I = "00" and ADR_10 = "00" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = WORD and SIZE_I = "00" and ADR_10 = "01" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = WORD and SIZE_I = "00" and ADR_10 = "10" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = WORD and SIZE_I = "00" and ADR_10 = "11" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = WORD and SIZE_I = "11" and ADR_10 = "00" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = WORD and SIZE_I = "11" and ADR_10 = "01" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = WORD and SIZE_I = "11" and ADR_10 = "10" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = WORD and SIZE_I = "11" and ADR_10 = "11" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = WORD and SIZE_I = "10" and ADR_10 = "00" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = WORD and SIZE_I = "10" and ADR_10 = "01" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = WORD and SIZE_I = "10" and ADR_10 = "10" then
					SIZE_I <= SIZE_I - "10";
				elsif BUS_WIDTH = WORD and SIZE_I = "10" and ADR_10 = "11" then
					SIZE_I <= SIZE_I - "01";
				elsif BUS_WIDTH = WORD and SIZE_I = "01" then
					SIZE_I <= SIZE_I - "01";
				--
				elsif BUS_WIDTH = BYTE then
					SIZE_I <= SIZE_I - "01";
				end if;
			end if;
		end if;
	end process P_SIZE;

	P_ADR_OFFS: process(RESETn, CLK)
	-- This process provides a temporary address offset during
	-- bus access over several bytes. The 6 bits are used for 
	-- the MOVEM command. other commands requires a maximum of
	-- 3 bits.
	begin
		if RESETn = '0' then
			ADR_OFFS_I <= "000000";
		elsif CLK = '1' and CLK' event then
			if ADR_OFFS_CLR = '1' then
				ADR_OFFS_I <= "000000";
			elsif BUS_CYC_RDY = '1' and HALTn = '0' and BERRn = '0' then
			 	null; -- Do not update if there is a retry cycle.
			elsif BUS_CYC_RDY = '1' then
				if MOVEP = '1' and OPCODE_FETCH = '0' then
					ADR_OFFS_I <= ADR_OFFS_I + "10";
				elsif BUS_WIDTH = LONG_32 and ADR_10 = "11" then
					ADR_OFFS_I <= ADR_OFFS_I + "01";
				elsif BUS_WIDTH = LONG_32 and ADR_10 = "10" then
					ADR_OFFS_I <= ADR_OFFS_I + "10";
				elsif BUS_WIDTH = LONG_32 and ADR_10 = "01" then
					ADR_OFFS_I <= ADR_OFFS_I + "11";
				elsif BUS_WIDTH = LONG_32 and ADR_10 = "00" then
					ADR_OFFS_I <= ADR_OFFS_I + "100";
				elsif BUS_WIDTH = WORD and ADR_10 = "11" then
					ADR_OFFS_I <= ADR_OFFS_I + "01";
				elsif BUS_WIDTH = WORD and ADR_10 = "10" then
					ADR_OFFS_I <= ADR_OFFS_I + "10";
				elsif BUS_WIDTH = WORD and ADR_10 = "01" then
					ADR_OFFS_I <= ADR_OFFS_I + "01";
				elsif BUS_WIDTH = WORD and ADR_10 = "00" then
					ADR_OFFS_I <= ADR_OFFS_I + "10";
				elsif BUS_WIDTH = BYTE then
					ADR_OFFS_I <= ADR_OFFS_I + "01";
				end if;
			end if;
		end if;
	end process P_ADR_OFFS;

	ADR_OFFSET <= ADR_OFFS_I;

	-- Progress controls:
	BUS_CTRL_RDY <= '1' when BUS_CTRL_STATE /= IDLE and NEXT_BUS_CTRL_STATE = IDLE else
					'1' when BUS_CTRL_STATE = IDLE and RD_BUS = '1' and OPCODE_FETCH = '1' and OPC_BUFF_STATUS = full else '0';
	BUS_CYC_RDY <= 	'1' when T_SLICE = S5 else '0';

	-- Bus tri state controls:
	DATA_PORT_EN <= '1' when DATA_EN = '1' and BUS_CTRL_STATE = BUSY_WR else '0';

	P_RMC: process(RESETn, CLK)
	-- This flip flop provides the information about the the
	-- read modify write lock during TAS, CAS and CAS2 operations.
	begin
		if RESETn = '0' then
			RMC_LOCK <= '0';
		elsif CLK = '1' and CLK' event then
			if RMC_OP = '1' and BUS_CTRL_STATE = IDLE and NEXT_BUS_CTRL_STATE /= IDLE then
				RMC_LOCK <= '1';
			elsif OP_END = '1' then
				RMC_LOCK <= '0';
			end if;
		end if;
	end process P_RMC;

	-- Bus control signals:
	ASn  <=	'0' when AS_EN = '1' else '1';
	DSn <= '0' when DS_RD_EN = '1' or DS_WR_EN = '1' else '1';
	RWn  <=	'0' when BUS_CTRL_STATE = BUSY_WR else '1';
	RMCn <= not RMC_LOCK;

	-- During read, write or read modify write processes, the read access is delayed by wait 
	-- states (slow read, slow write) if there is no DSACKn asserted until the end of S3.
	-- In case of the autovector interrupt acknowledge cycle, the bus cycle is finished by
	-- the AVECn signal and in case of no response, the BERRn finishes the bus cycle.
	-- If the bus cycle is aborted, it terminates as if BERRn occured.
	WAITSTATES <= DSACKn(1) and DSACKn(0) and AVECn and BERRn and not EXEC_ABORT;

	P_BUSWIDTH: process
	-- This process provides the information about the width of the bus at the selected
	-- address. This information is sampled during bus state S3. The implementation of the bus
	-- logic accepts operand memory of 8, 16 and 32 bit bus width. The information of the bus
	-- width is sampled during DSACKn is active and hold during DSACKn is deasserted.
	begin
		wait until CLK = '1' and CLK' event;
		case DSACKn is
			when "00" => BUS_WIDTH <= LONG_32;
			when "01" => BUS_WIDTH <= WORD;
			when "10" => BUS_WIDTH <= BYTE;
			when "11" => null; -- Hold.
		end case;
	end process P_BUSWIDTH;

	SLICES: process(RESETn, CLK)
	-- This process provides the central timing for the read, write and read modify write cycle as also
	-- for the bus arbitration procedure.
	begin
		if RESETn = '0' then
			SLICE_CNT <= "11";
		elsif CLK = '1' and CLK' event then
			if BUS_CTRL_STATE = IDLE then
				SLICE_CNT <= "11"; -- Init.
			-- Normal counting:
			elsif SLICE_CNT = "10" and NEXT_BUS_CTRL_STATE /= IDLE then
				SLICE_CNT <= "00"; -- Start directly the next read or write cycle.
			elsif SLICE_CNT = "10" then
				SLICE_CNT <= "11"; -- Finish the read or write cycle.
			elsif NEXT_BUS_CTRL_STATE /= IDLE and WAITSTATES = '0' then
				SLICE_CNT <= SLICE_CNT + '1'; -- Cycle active.
			end if;
		end if;
	end process SLICES;

	T_SLICE <=  S0 when SLICE_CNT = "00" and CLK = '1' else
			    S1 when SLICE_CNT = "00" and CLK = '0' else
				S2 when WAITSTATES = '0' and SLICE_CNT = "01" and CLK = '1' else
				S3 when WAITSTATES = '1' and SLICE_CNT = "01" and CLK = '1' else
				S3 when SLICE_CNT = "01" and CLK = '0' else
				S4 when SLICE_CNT = "10" and CLK = '1' else
			    S5 when SLICE_CNT = "10" and CLK = '0' else IDLE;

	-- The modelling with the two processes working on the positive and negative clock edge
	-- is a bit complicated. But it results in rather 'clean' (glitch free) bus control
	-- signals. Every signal is modelled with it's own timing to give the core a high degree
	-- of freedom.
	N: process
	begin
		wait until CLK = '0' and CLK' event;
		case T_SLICE is
			when S2 => DS_WR_EN_N <= '1';
			when others => DS_WR_EN_N <= '0';
		end case;	
		case T_SLICE is
			when S0 | S2 => DS_RD_EN_N <= '1';
			when others => DS_RD_EN_N <= '0';
		end case;	
		case T_SLICE is
			when S0 | S2 => AS_EN_N <= '1';
			when others => AS_EN_N <= '0';
		end case;	
	end process N;

	P: process
	begin
		wait until CLK = '1' and CLK' event;
		case T_SLICE is
			when S1 | S3 => DATA_EN_P <= '1';
			when others => DATA_EN_P <= '0';
		end case;	
	end process P;

	-- Read and write timing:
	DS_WR_EN <= DS_WR_EN_N; 	-- '1' when S3 | S4.
	DS_RD_EN <= DS_RD_EN_N; 	-- '1' when S1 | S2 | S3 | S4.
	AS_EN <= AS_EN_N; 			-- '1' when S1 | S2 | S3 | S4.
	DATA_EN <= DATA_EN_P; 		-- '1' when S2 | S3 | S4 | S5.

	-- Timing valid for all modes:
	BUS_EN <= BUS_EN_I;
	BUS_EN_I <= '0' when BGACKn = '0' else '1'; -- During arbitration.
	
	-- Bus arbitration:
	ARB_REG: process(RESETn, CLK)
	begin
		if RESETn = '0' then
			ARB_STATE <= IDLE;
		elsif CLK = '1' and CLK' event then
			ARB_STATE <= NEXT_ARB_STATE;
		end if;
	end process ARB_REG;
	
	ARB_DEC: process(ARB_STATE, BRn, BUS_CTRL_STATE, T_SLICE, BGACKn, RMC_LOCK)
	-- This is the bus arbitration state machine. Arbitration takes place, when the bus
	-- is not used by the CPU.
	begin
		case ARB_STATE is
			when IDLE =>
				if BRn = '0' and BUS_CTRL_STATE = IDLE and RMC_LOCK = '0' then
					NEXT_ARB_STATE <= GRANT;
				else
					NEXT_ARB_STATE <= IDLE;
				end if;
			when GRANT =>
				if BGACKn = '0' then
					NEXT_ARB_STATE <= WAIT_RELEASE_2WIRE;
				elsif BRn = '1' then
					NEXT_ARB_STATE <= IDLE; -- Re-initialize.
				else
					NEXT_ARB_STATE <= GRANT;
				end if;
			when WAIT_RELEASE_2WIRE =>
				-- In the case of a 2 wire bus arbitration, no re-entry for another
				-- device is possible.
				if BGACKn = '1' and BRn = '1' then
					NEXT_ARB_STATE <= IDLE;
				elsif BGACKn = '0' then
					NEXT_ARB_STATE <= WAIT_RELEASE_3WIRE; -- BGACKn = '0' indicates 3 wire arbitration.
				else
					NEXT_ARB_STATE <= WAIT_RELEASE_2WIRE;
				end if;
			when WAIT_RELEASE_3WIRE =>
				if BGACKn = '1' and BRn = '0' then
					NEXT_ARB_STATE <= GRANT; -- Re-enter new arbitration.
				elsif BGACKn = '1' then
					NEXT_ARB_STATE <= IDLE;
				else
					NEXT_ARB_STATE <= WAIT_RELEASE_3WIRE;
				end if;
		end case;
	end process ARB_DEC;

	BGn <= 	'0' when ARB_STATE = GRANT else -- 3 wire and 2 wire arbitration.
			'0' when ARB_STATE = WAIT_RELEASE_2WIRE else '1'; -- 2 wire arbitration.

	BUS_LOCKED <= 	'1' when ARB_STATE = GRANT and T_SLICE = IDLE else
					 -- S6 locking due to special case:
					'1' when ARB_STATE = WAIT_RELEASE_2WIRE and T_SLICE /= S4 else
					'1' when ARB_STATE = WAIT_RELEASE_3WIRE else '0';

	-- RESET logic:
	RESET_FILTER: process(RESETn, CLK)
	-- This process filters the incoming reset pin.
	-- If RESET_INn and HALTn are asserted together for longer
	-- than 10 clock cycles over the execution of a CPU reset
	-- command, the CPU reset is released.
	variable TMP	: std_logic_vector(3 downto 0);
	begin
		if RESETn = '0' then
			TMP := x"F"; -- For correct startup.
			RESET_CPU_In <= '0';
		elsif CLK = '1' and CLK' event then
			if RESET_INn = '0' and HALTn = '0' and RESET_OUT_EN_I = '0' and TMP < x"F" then
				TMP := TMP + '1';
			elsif RESET_INn = '1' or HALTn = '1' or RESET_OUT_EN_I = '1' then
				TMP := x"0";
			end if;
			if TMP > x"A" then
				RESET_CPU_In <= '0'; -- Release internal reset.
			else
				RESET_CPU_In <= '1';
			end if;
		end if;
	end process RESET_FILTER;

	RESET_TIMER: process(RESETn, CLK)
	-- This logic is responsible for the assertion of the
	-- reset output for 512 clock cycles, during the reset
	-- command. The LOCK variable avoids re-initialisation
	-- of the counter in the case that the RESET_EN is no
	-- strobe.
	variable TMP	: std_logic_vector(8 downto 0);
	variable LOCK	: boolean;
	begin
		if RESETn = '0' then
			TMP := (others => '0');
			LOCK := false;
		elsif CLK = '1' and CLK' event then
			if RESET_EN = '1' and LOCK = false then
				TMP := "111111111"; -- 512 initial value.
				LOCK := true; -- Lock the counter initialisation.
			elsif TMP > "000000000" then
				TMP := TMP - '1';
			elsif RESET_EN = '0' then
				LOCK := false; -- Unlock the counter initialisation.
			end if;
			if TMP = "000000001" then 
				RESET_OUT_EN_I <= '0';
				RESET_RDY <= '1';
			elsif TMP <= "000000011" then
				RESET_OUT_EN_I <= '0';
				RESET_RDY <= '0';
			else
				RESET_OUT_EN_I <= '1';
				RESET_RDY <= '0';
			end if;
		end if;
	end process RESET_TIMER;

	RESET_OUT_EN <= RESET_OUT_EN_I;
	RESET_CPUn <= RESET_CPU_In;
end BEHAVIOR;
