----------------------------------------------------------------------
----                                                              ----
---- MC68K family compatible IP Core                              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- This file is the top level file of this ip core.             ----
----                                                              ----
----                                                              ----
---- Description:                                                 ----
---- This model provides a MC68851 compatible memory management   ----
---- (MMU) core.                                                  ----
----                                                              ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright (C) 2007 Wolfgang Foerster                         ----
----                                                              ----
---- This source file may be used and distributed without         ----
---- restriction provided that this copyright statement is not    ----
---- removed from the file and that any derivative work contains  ----
---- the original copyright notice and the associated disclaimer. ----
----                                                              ----
---- This source file is free software; you can redistribute it   ----
---- and/or modify it under the terms of the GNU Lesser General   ----
---- Public License as published by the Free Software Foundation; ----
---- either version 2.1 of the License, or (at your option) any   ----
---- later version.                                               ----
----                                                              ----
---- This source is distributed in the hope that it will be       ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied   ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ----
---- PURPOSE. See the GNU Lesser General Public License for more  ----
---- details.                                                     ----
----                                                              ----
---- You should have received a copy of the GNU Lesser General    ----
---- Public License along with this source; if not, download it   ----
---- from http://www.gnu.org/licenses/lgpl.html                   ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6B  2006/12/24 WF
--   Initial Release.

library work;
use work.wf68851ip_pkg.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WF68851IP_TOP is
	port (
		CLK				: in bit;
		RESET_COREn		: in bit; -- Core reset.
		
		-- System control:
		RESETn			: in bit;
		ASYNCn			: in bit;
		CLIn			: out bit;
		BERRn_IN		: in bit;
		BERRn_OUT		: out bit;
		BERRn_EN		: out bit;
		HALTn_IN		: in bit;
		HALTn_OUT		: out bit;
		HALTn_EN		: out bit;
		
		-- Bus status:
		FC_IN			: in bit_vector(3 downto 0);
		FC_OUT			: out bit_vector(3 downto 0);
		FC_EN			: out bit;
		
		-- Address:
		ADR_IN			: in bit_vector(31 downto 0);
		ADR_OUT			: out bit_vector(31 downto 0);
		ADR_EN			: out bit;
		LADR			: in bit_vector(31 downto 8);
		PADR			: out bit_vector(31 downto 8);

		-- Data:
		DATA_IN			: in std_logic_vector(31 downto 0);
		DATA_OUT		: in std_logic_vector(31 downto 0);
		DATA_EN			: out bit;

		-- Aynchronous bus control:
		SIZE_IN			: in std_logic_vector(1 downto 0);
		SIZE_OUT		: out std_logic_vector(1 downto 0);
		SIZE_EN			: out std_logic;
		DSACKn_IN		: in bit_vector(1 downto 0);
		DSACKn_OUT		: out bit_vector(1 downto 0);
		DSACKn_EN		: out bit;
		LASn			: in std_logic;
		PASn			: out std_logic;
		RMCn_IN			: in std_logic;
		RMCn_OUT		: out std_logic;
		RMCn_EN			: out std_logic;
		RWn_IN			: in std_logic;
		RWn_OUT			: out std_logic;
		RWn_EN			: out std_logic;
		DSn_IN			: in std_logic;
		DSn_OUT			: out std_logic;
		DSn_EN			: out std_logic;
		DBDIS			: out bit;

		-- Bus arbitration control:
		PBRn			: in bit;
		PBGn			: out bit;
		PBGACKn			: in bit;
		LBRIn			: in bit;
		LBROn			: out bit;
		LBGIn			: out bit;
		LBGOn			: out bit;
		LBGACKn_IN		: in bit;
		LBGACKn_OUT		: out bit;
		LBGACKn_EN		: out bit
	);
end entity WF68851IP_TOP;
	
architecture STRUCTURE of WF68851IP_TOP is
begin
SIGNAL_SAMPLE: process
	-- The bus control signals used in this core are sampled on the negative clock
	-- edge. Thus the signals are valid on the following positive clock edge. In the original
	-- 68K machines, the input synchronisation is realized with three latches (see the 68K
	-- user manual for more information). This concept is not suitable for a FPGA design 
	-- and therefore not used here.
	begin
		wait until CLK = '0' and CLK' event;
	end process SIGNAL_SAMPLE;

	I_BUS_IF: WF68851IP_BUS_INTERFACE
	port map(
		CLK					=> CLK,
		RESETn				=> RESET_COREn,
		RESET_INn			=> RESET_IN_In,
		RESET_OUT_EN		=> RESET_OUT_EN_I,
		RESET_CPUn			=> RESET_CPU_In,
		RESET_EN			=> RESET_EN_I,
		RESET_RDY			=> RESET_RDY_I,
		DATA_PORT_IN		=> DATA_PORT_IN,
		DATA_PORT_OUT		=> DATA_PORT_OUT,
		DATA_2_PORT			=> DATA_2_PORT,
		DATA_2_CORE			=> DATA_2_CORE,
		OPCODE_DATA			=> OPCODE_DATA_I,
		OP_SIZE				=> OP_SIZE_BUS,
		OP_END				=> OP_END_I,
		RMC_OP				=> RMC_OP_I,
		SEL_BUFF_A			=> SEL_BUFF_A_I,
		SEL_BUFF_B			=> SEL_BUFF_B_I,
		BUFFER_A			=> BUS_BUFFER_A,
		BUFFER_B			=> BUS_BUFFER_B,
		RD_BUS				=> RD_BUS_I,
		WR_BUS				=> WR_BUS_I,
		A0					=> ADR_PORT(0),
		A1					=> ADR_PORT(1),
		PC1					=> PC_OUT(1),
		EXEC_ABORT			=> EXEC_ABORT_I,
		ADR_OFFS_CLR		=> ADR_OFFS_CLR_I,
		MOVEP				=> MOVEP_I,
		ADR_OFFSET			=> ADR_OFFSET_I,
		BUS_CTRL_RDY		=> BUS_CTRL_RDY_I,
		DATA_VALID			=> DATA_VALID_I,
		DATA_PORT_EN		=> DATA_PORT_EN_I,
		BUS_EN				=> BUS_EN_I,
		DSACKn				=> DSACK_In,
		BERRn				=> BERR_In,
		AVECn				=> AVEC_In,
		HALTn				=> HALT_In,
		SIZE				=> SIZE_I,
		ASn					=> AS_In,
		DSn					=> DS_In,
		RWn					=> RWn_I,
		RMCn				=> RMC_In,
		BRn					=> BR_In,
		BGACKn				=> BGACK_In,
		BGn					=> BGn,
		OPCODE_FETCH		=> OPCODE_FETCH_I,
		FLUSH_OPC_BUFF		=> FLUSH_OPC_BUFF_I
	);
end STRUCTURE;
